@@ -20,7 +20,6 @@ import uuid
 
 from oslo_config import cfg
 from oslo_log import log
-import retrying
 import six
 from six.moves.urllib.parse import quote
 try:
@@ -31,6 +30,7 @@ except ImportError:
 
 from gnocchi import storage
 from gnocchi.storage import _carbonara
+from gnocchi import utils
 
 LOG = log.getLogger(__name__)
 
@@ -49,12 +49,19 @@ OPTS = [
     cfg.StrOpt('swift_user',
                default="admin:admin",
                help='Swift user.'),
+    cfg.StrOpt('swift_user_domain_name',
+               default='Default',
+               help='Swift user domain name.'),
     cfg.StrOpt('swift_key',
                secret=True,
                default="admin",
                help='Swift key/password.'),
-    cfg.StrOpt('swift_tenant_name',
-               help='Swift tenant name, only used in v2 auth.'),
+    cfg.StrOpt('swift_project_name',
+               help='Swift tenant name, only used in v2/v3 auth.',
+               deprecated_name="swift_tenant_name"),
+    cfg.StrOpt('swift_project_domain_name',
+               default='Default',
+               help='Swift project domain name.'),
     cfg.StrOpt('swift_container_prefix',
                default='gnocchi',
                help='Prefix to namespace metric containers.'),
@@ -68,37 +75,40 @@ OPTS = [
 ]
 
 
-def retry_if_result_empty(result):
-    return len(result) == 0
-
-
 class SwiftStorage(_carbonara.CarbonaraBasedStorage):
 
+    WRITE_FULL = True
     POST_HEADERS = {'Accept': 'application/json', 'Content-Type': 'text/plain'}
 
     def __init__(self, conf):
         super(SwiftStorage, self).__init__(conf)
         if swclient is None:
             raise RuntimeError("python-swiftclient unavailable")
-        self.swift = swclient.Connection(
+        self.swift = self._get_connection(conf)
+        self._container_prefix = conf.swift_container_prefix
+        self.swift.put_container(self.MEASURE_PREFIX)
+
+    @utils.retry
+    def _get_connection(self, conf):
+        return swclient.Connection(
             auth_version=conf.swift_auth_version,
             authurl=conf.swift_authurl,
             preauthtoken=conf.swift_preauthtoken,
             user=conf.swift_user,
             key=conf.swift_key,
-            tenant_name=conf.swift_tenant_name,
+            tenant_name=conf.swift_project_name,
             timeout=conf.swift_timeout,
-            os_options={'endpoint_type': conf.swift_endpoint_type},
-            retries=1)
-        self._container_prefix = conf.swift_container_prefix
-        self.swift.put_container(self.MEASURE_PREFIX)
+            os_options={'endpoint_type': conf.swift_endpoint_type,
+                        'user_domain_name': conf.swift_user_domain_name},
+            retries=0)
 
     def _container_name(self, metric):
         return '%s.%s' % (self._container_prefix, str(metric.id))
 
     @staticmethod
-    def _object_name(split_key, aggregation, granularity):
-        return '%s_%s_%s' % (split_key, aggregation, granularity)
+    def _object_name(split_key, aggregation, granularity, version=3):
+        name = '%s_%s_%s' % (split_key, aggregation, granularity)
+        return name + '_v%s' % version if version else name
 
     def _create_metric(self, metric):
         # TODO(jd) A container per user in their account?
@@ -118,19 +128,23 @@ class SwiftStorage(_carbonara.CarbonaraBasedStorage):
             data)
 
     def _build_report(self, details):
-        headers, files = self.swift.get_container(self.MEASURE_PREFIX,
-                                                  delimiter='/',
-                                                  full_listing=True)
-        metrics = len(files)
-        measures = int(headers.get('x-container-object-count'))
         metric_details = defaultdict(int)
         if details:
             headers, files = self.swift.get_container(self.MEASURE_PREFIX,
                                                       full_listing=True)
+            metrics = set()
             for f in files:
-                metric = f['name'].split('/', 1)[0]
+                metric, metric_files = f['name'].split("/", 1)
                 metric_details[metric] += 1
-        return metrics, measures, metric_details if details else None
+                metrics.add(metric)
+            nb_metrics = len(metrics)
+        else:
+            headers, files = self.swift.get_container(self.MEASURE_PREFIX,
+                                                      delimiter='/',
+                                                      full_listing=True)
+            nb_metrics = len(files)
+        measures = int(headers.get('x-container-object-count'))
+        return nb_metrics, measures, metric_details if details else None
 
     def list_metric_with_measures_to_process(self, size, part, full=False):
         limit = None
@@ -180,25 +194,27 @@ class SwiftStorage(_carbonara.CarbonaraBasedStorage):
         for f in files:
             headers, data = self.swift.get_object(
                 self.MEASURE_PREFIX, f['name'])
-            measures.extend(self._unserialize_measures(data))
+            measures.extend(self._unserialize_measures(f['name'], data))
 
         yield measures
 
         # Now clean objects
         self._bulk_delete(self.MEASURE_PREFIX, files)
 
-    def _store_metric_measures(self, metric, timestamp_key,
-                               aggregation, granularity, data):
+    def _store_metric_measures(self, metric, timestamp_key, aggregation,
+                               granularity, data, offset=None, version=3):
         self.swift.put_object(
             self._container_name(metric),
-            self._object_name(timestamp_key, aggregation, granularity),
+            self._object_name(timestamp_key, aggregation, granularity,
+                              version),
             data)
 
     def _delete_metric_measures(self, metric, timestamp_key, aggregation,
-                                granularity):
+                                granularity, version=3):
         self.swift.delete_object(
             self._container_name(metric),
-            self._object_name(timestamp_key, aggregation, granularity))
+            self._object_name(timestamp_key, aggregation, granularity,
+                              version))
 
     def _delete_metric(self, metric):
         self._delete_unaggregated_timeserie(metric)
@@ -219,14 +235,12 @@ class SwiftStorage(_carbonara.CarbonaraBasedStorage):
                     # Deleted in the meantime? Whatever.
                     raise
 
-    @retrying.retry(stop_max_attempt_number=4,
-                    wait_fixed=500,
-                    retry_on_result=retry_if_result_empty)
-    def _get_measures(self, metric, timestamp_key, aggregation, granularity):
+    def _get_measures(self, metric, timestamp_key, aggregation, granularity,
+                      version=3):
         try:
             headers, contents = self.swift.get_object(
                 self._container_name(metric), self._object_name(
-                    timestamp_key, aggregation, granularity))
+                    timestamp_key, aggregation, granularity, version))
         except swclient.ClientException as e:
             if e.http_status == 404:
                 try:
@@ -239,7 +253,8 @@ class SwiftStorage(_carbonara.CarbonaraBasedStorage):
             raise
         return contents
 
-    def _list_split_keys_for_metric(self, metric, aggregation, granularity):
+    def _list_split_keys_for_metric(self, metric, aggregation, granularity,
+                                    version=None):
         container = self._container_name(metric)
         try:
             headers, files = self.swift.get_container(
@@ -248,61 +263,43 @@ class SwiftStorage(_carbonara.CarbonaraBasedStorage):
             if e.http_status == 404:
                 raise storage.MetricDoesNotExist(metric)
             raise
-        keys = []
+        keys = set()
         for f in files:
             try:
-                key, agg, g = f['name'].split('_', 2)
-            except ValueError:
+                meta = f['name'].split('_')
+                if (aggregation == meta[1] and granularity == float(meta[2])
+                        and self._version_check(f['name'], version)):
+                    keys.add(meta[0])
+            except (ValueError, IndexError):
                 # Might be "none", or any other file. Be resilient.
                 continue
-            if aggregation == agg and granularity == float(g):
-                keys.append(key)
         return keys
 
-    @retrying.retry(stop_max_attempt_number=4,
-                    wait_fixed=500,
-                    retry_on_result=retry_if_result_empty)
-    def _get_unaggregated_timeserie(self, metric):
+    @staticmethod
+    def _build_unaggregated_timeserie_path(version):
+        return 'none' + ("_v%s" % version if version else "")
+
+    def _get_unaggregated_timeserie(self, metric, version=3):
         try:
             headers, contents = self.swift.get_object(
-                self._container_name(metric), "none")
+                self._container_name(metric),
+                self._build_unaggregated_timeserie_path(version))
         except swclient.ClientException as e:
             if e.http_status == 404:
                 raise storage.MetricDoesNotExist(metric)
             raise
         return contents
 
-    def _store_unaggregated_timeserie(self, metric, data):
-        self.swift.put_object(self._container_name(metric), "none", data)
+    def _store_unaggregated_timeserie(self, metric, data, version=3):
+        self.swift.put_object(self._container_name(metric),
+                              self._build_unaggregated_timeserie_path(version),
+                              data)
 
-    def _delete_unaggregated_timeserie(self, metric):
+    def _delete_unaggregated_timeserie(self, metric, version=3):
         try:
-            self.swift.delete_object(self._container_name(metric), "none")
+            self.swift.delete_object(
+                self._container_name(metric),
+                self._build_unaggregated_timeserie_path(version))
         except swclient.ClientException as e:
             if e.http_status != 404:
                 raise
-
-    # The following methods deal with Gnocchi <= 1.3 archives
-    def _get_metric_archive(self, metric, aggregation):
-        """Retrieve data in the place we used to store TimeSerieArchive."""
-        try:
-            headers, contents = self.swift.get_object(
-                self._container_name(metric), aggregation)
-        except swclient.ClientException as e:
-            if e.http_status == 404:
-                raise storage.AggregationDoesNotExist(metric, aggregation)
-            raise
-        return contents
-
-    def _store_metric_archive(self, metric, aggregation, data):
-        """Stores data in the place we used to store TimeSerieArchive."""
-        self.swift.put_object(self._container_name(metric), aggregation, data)
-
-    def _delete_metric_archives(self, metric):
-        for aggregation in metric.archive_policy.aggregation_methods:
-            try:
-                self.swift.delete_object(self._container_name(metric),
-                                         aggregation)
-            except swclient.ClientException as e:
-                if e.http_status != 404:
-                    raise
